@@ -14,22 +14,29 @@
 //! - `vtable_cache`: Enables local vtable caching to speed up lookups after the first. This
 //!   requires mutating the archive, which is not possible for all use cases.
 
+#![deny(broken_intra_doc_links)]
+#![deny(missing_docs)]
+#![deny(missing_crate_level_docs)]
 #![cfg_attr(feature = "nightly", feature(core_intrinsics))]
 
 #[cfg(feature = "validation")]
 pub mod validation;
 
 #[cfg(feature = "vtable_cache")]
-use core::sync::atomic::{AtomicU64, Ordering};
+use core::sync::atomic::AtomicU64;
 use core::{
-    alloc,
+    alloc::Layout,
     any::Any,
     hash::{Hash, Hasher},
     marker::PhantomData,
-    mem::MaybeUninit,
+    ptr,
 };
 use ptr_meta::{DynMetadata, Pointee};
-use rkyv::{de::Deserializer, project_struct, ser::Serializer, Fallible, Serialize};
+use rkyv::{
+    from_archived,
+    ser::{ScratchSpace, Serializer},
+    to_archived, Archived, Fallible, Serialize,
+};
 pub use rkyv_dyn_derive::archive_dyn;
 use rkyv_typename::TypeName;
 use std::collections::{hash_map::DefaultHasher, HashMap};
@@ -61,6 +68,25 @@ pub trait DynSerializer {
 
     /// Attempts to write the given bytes to the serializer.
     fn write_dyn(&mut self, bytes: &[u8]) -> Result<(), DynError>;
+
+    /// Allocates scratch space of the requested size.
+    ///
+    /// # Safety
+    ///
+    /// `layout` must have non-zero size.
+    unsafe fn push_scratch_dyn(&mut self, layout: Layout) -> Result<ptr::NonNull<[u8]>, DynError>;
+
+    /// Deallocates previously allocated scratch space.
+    ///
+    /// # Safety
+    ///
+    /// - `ptr` must be the scratch memory last allocated with `push_scratch`.
+    /// - `layout` must be the same layout that was used to allocate that block of memory.
+    unsafe fn pop_scratch_dyn(
+        &mut self,
+        ptr: ptr::NonNull<u8>,
+        layout: Layout,
+    ) -> Result<(), DynError>;
 }
 
 impl<'a> Fallible for dyn DynSerializer + 'a {
@@ -77,16 +103,41 @@ impl<'a> Serializer for dyn DynSerializer + 'a {
     }
 }
 
-impl<S: Serializer + ?Sized> DynSerializer for &mut S {
+impl<'a> ScratchSpace for dyn DynSerializer + 'a {
+    unsafe fn push_scratch(&mut self, layout: Layout) -> Result<ptr::NonNull<[u8]>, Self::Error> {
+        self.push_scratch_dyn(layout)
+    }
+
+    unsafe fn pop_scratch(
+        &mut self,
+        ptr: ptr::NonNull<u8>,
+        layout: Layout,
+    ) -> Result<(), Self::Error> {
+        self.pop_scratch_dyn(ptr, layout)
+    }
+}
+
+impl<S: ScratchSpace + Serializer + ?Sized> DynSerializer for &mut S {
     fn pos_dyn(&self) -> usize {
         self.pos()
     }
 
     fn write_dyn(&mut self, bytes: &[u8]) -> Result<(), DynError> {
-        match self.write(bytes) {
-            Ok(()) => Ok(()),
-            Err(e) => Err(Box::new(e)),
-        }
+        self.write(bytes).map_err(|e| Box::new(e) as DynError)
+    }
+
+    unsafe fn push_scratch_dyn(&mut self, layout: Layout) -> Result<ptr::NonNull<[u8]>, DynError> {
+        self.push_scratch(layout)
+            .map_err(|e| Box::new(e) as DynError)
+    }
+
+    unsafe fn pop_scratch_dyn(
+        &mut self,
+        ptr: ptr::NonNull<u8>,
+        layout: Layout,
+    ) -> Result<(), DynError> {
+        self.pop_scratch(ptr, layout)
+            .map_err(|e| Box::new(e) as DynError)
     }
 }
 
@@ -118,22 +169,22 @@ fn hash_type<T: TypeName + ?Sized>() -> u64 {
 /// are boxed as regular trait objects. This is because your deserialized values have to implement
 /// `SerializeDyn` but your archived values do not.
 ///
-/// ## Examples
+/// # Examples
 ///
 /// See [`archive_dyn`](macro@archive_dyn) for customization options.
 ///
 /// ```
 /// use rkyv::{
 ///     archived_value,
-///     de::deserializers::AllocDeserializer,
 ///     ser::{
-///         serializers::AlignedSerializer,
+///         serializers::AllocSerializer,
 ///         Serializer,
 ///     },
 ///     AlignedVec,
 ///     Archive,
 ///     Archived,
 ///     Deserialize,
+///     Infallible,
 ///     Serialize,
 /// };
 /// use rkyv_dyn::archive_dyn;
@@ -145,7 +196,7 @@ fn hash_type<T: TypeName + ?Sized>() -> u64 {
 /// }
 ///
 /// #[derive(Archive, Serialize, Deserialize)]
-/// #[archive(derive(TypeName))]
+/// #[archive_attr(derive(TypeName))]
 /// struct StringStruct(String);
 ///
 /// #[archive_dyn(deserialize)]
@@ -162,7 +213,7 @@ fn hash_type<T: TypeName + ?Sized>() -> u64 {
 /// }
 ///
 /// #[derive(Archive, Serialize, Deserialize)]
-/// #[archive(derive(TypeName))]
+/// #[archive_attr(derive(TypeName))]
 /// struct IntStruct(i32);
 ///
 /// #[archive_dyn(deserialize)]
@@ -181,12 +232,12 @@ fn hash_type<T: TypeName + ?Sized>() -> u64 {
 /// let boxed_int = Box::new(IntStruct(42)) as Box<dyn SerializeExampleTrait>;
 /// let boxed_string = Box::new(StringStruct("hello world".to_string()))
 ///     as Box<dyn SerializeExampleTrait>;
-/// let mut serializer = AlignedSerializer::new(AlignedVec::new());
+/// let mut serializer = AllocSerializer::<256>::default();
 /// let int_pos = serializer.serialize_value(&boxed_int)
 ///     .expect("failed to archive boxed int");
 /// let str_pos = serializer.serialize_value(&boxed_string)
 ///     .expect("failed to archive boxed string");
-/// let buf = serializer.into_inner();
+/// let buf = serializer.into_serializer().into_inner();
 /// let archived_int = unsafe {
 ///     archived_value::<Box<dyn SerializeExampleTrait>>(buf.as_ref(), int_pos)
 /// };
@@ -197,9 +248,9 @@ fn hash_type<T: TypeName + ?Sized>() -> u64 {
 /// assert_eq!(archived_string.value(), "hello world");
 ///
 /// let deserialized_int: Box<dyn SerializeExampleTrait> = archived_int
-///     .deserialize(&mut AllocDeserializer).unwrap();
+///     .deserialize(&mut Infallible).unwrap();
 /// let deserialized_string: Box<dyn SerializeExampleTrait> = archived_string
-///     .deserialize(&mut AllocDeserializer).unwrap();
+///     .deserialize(&mut Infallible).unwrap();
 /// assert_eq!(deserialized_int.value(), "42");
 /// assert_eq!(deserialized_string.value(), "hello world");
 /// ```
@@ -225,31 +276,13 @@ where
 }
 
 /// An object-safe version of `Deserializer`.
-pub trait DynDeserializer {
-    /// Allocates and returns memory with the given layout.
-    ///
-    /// # Safety
-    ///
-    /// The caller must guarantee that the memory returned by this function is deallocated by the
-    /// global allocator.
-    unsafe fn alloc_dyn(&mut self, layout: alloc::Layout) -> Result<*mut u8, DynError>;
-}
+pub trait DynDeserializer {}
 
 impl<'a> Fallible for dyn DynDeserializer + 'a {
     type Error = DynError;
 }
 
-impl<'a> Deserializer for (dyn DynDeserializer + 'a) {
-    unsafe fn alloc(&mut self, layout: alloc::Layout) -> Result<*mut u8, Self::Error> {
-        self.alloc_dyn(layout)
-    }
-}
-
-impl<D: Deserializer + ?Sized> DynDeserializer for &mut D {
-    unsafe fn alloc_dyn(&mut self, layout: alloc::Layout) -> Result<*mut u8, DynError> {
-        self.alloc(layout).map_err(|e| Box::new(e) as DynError)
-    }
-}
+impl<D: Fallible + ?Sized> DynDeserializer for &mut D {}
 
 /// A trait object that can be deserialized.
 ///
@@ -259,10 +292,11 @@ pub trait DeserializeDyn<T: Pointee + ?Sized> {
     ///
     /// # Safety
     ///
-    /// The caller must guarantee that the memory returned is properly deallocated.
+    /// The memory returned must be properly deallocated.
     unsafe fn deserialize_dyn(
         &self,
         deserializer: &mut dyn DynDeserializer,
+        alloc: &mut dyn FnMut(Layout) -> *mut u8,
     ) -> Result<*mut (), DynError>;
 
     /// Returns the metadata for the deserialized version of this value.
@@ -275,34 +309,32 @@ pub trait DeserializeDyn<T: Pointee + ?Sized> {
 /// The archived version of `DynMetadata`.
 #[cfg_attr(feature = "strict", repr(C))]
 pub struct ArchivedDynMetadata<T: ?Sized> {
-    type_id: u64,
+    type_id: Archived<u64>,
     #[cfg(feature = "vtable_cache")]
-    cached_vtable: AtomicU64,
+    cached_vtable: Archived<AtomicU64>,
     #[cfg(not(feature = "vtable_cache"))]
-    cached_vtable: u64,
+    #[allow(dead_code)]
+    cached_vtable: Archived<u64>,
     phantom: PhantomData<T>,
 }
 
 impl<T: TypeName + ?Sized> ArchivedDynMetadata<T> {
     /// Creates a new `ArchivedDynMetadata` for the given type.
-    pub fn emplace(type_id: u64, out: &mut MaybeUninit<Self>) {
-        unsafe {
-            project_struct!(out: Self => type_id: u64)
-                .as_mut_ptr()
-                .write(type_id);
-            #[cfg(feature = "vtable_cache")]
-            (&mut *project_struct!(out: Self => cached_vtable: AtomicU64).as_mut_ptr())
-                .store(0, Ordering::Relaxed);
-            #[cfg(not(feature = "vtable_cache"))]
-            project_struct!(out: Self => cached_vtable: u64)
-                .as_mut_ptr()
-                .write(0);
-        }
+    ///
+    /// # Safety
+    ///
+    /// `out` must point to a valid location for an `ArchivedDynMetadata<T>`.
+    pub unsafe fn emplace(type_id: u64, out: *mut Self) {
+        ptr::addr_of_mut!((*out).type_id).write(to_archived!(type_id));
+        #[cfg(feature = "vtable_cache")]
+        ptr::addr_of_mut!((*out).cached_vtable).write(Archived::<AtomicU64>::from(0u64));
+        #[cfg(not(feature = "vtable_cache"))]
+        ptr::addr_of_mut!((*out).cached_vtable).write(from_archived!(0u64));
     }
 
     fn lookup_vtable(&self) -> usize {
         IMPL_REGISTRY
-            .get::<T>(self.type_id)
+            .get::<T>(from_archived!(self.type_id))
             .expect("attempted to get vtable for an unregistered impl")
             .vtable
     }
@@ -405,6 +437,7 @@ pub struct ImplEntry {
 }
 
 impl ImplEntry {
+    #[doc(hidden)]
     pub fn new<TY: TypeName + RegisteredImpl<TR>, TR: TypeName + ?Sized>() -> Self {
         Self {
             impl_id: ImplId::new::<TY, TR>(),
@@ -491,7 +524,6 @@ macro_rules! register_validation {
 macro_rules! register_impl {
     ($type:ty as $trait:ty) => {
         const _: () = {
-            use core::mem::MaybeUninit;
             use rkyv_dyn::{
                 debug_info, inventory, register_validation, ImplData, ImplDebugInfo, ImplEntry,
                 RegisteredImpl,
